# coding: utf-8

import os
import re
import glob
import ssl
import datetime
import socket
import subprocess as sp

import pyquery
import click
import requests
from requests.auth import HTTPBasicAuth
from OpenSSL import crypto

BUFFER_DAYS = 25
NGINX_ENABLED_DIR = '/etc/nginx/sites-enabled'
LETSENCRYPT_CONF_DIR = '/etc/letsencrypt'

@click.command('domain')
@click.option('--ssl/--no-ssl', is_flag=True, help="Test/Don't test https", default=True)
@click.option('--www/--no-www', is_flag=True, help="Test/Don't prepend www", default=True)
@click.option('--user', '-u', help="The user for basic authentication")
@click.option('--password', '-p', help="The pass for basic authentication")
@click.option('--permutate', '-x', help="Permutate on all domains in string eg: -x 'it com org'")
@click.option('--verbose', '-v', help="Be verbose and declare the hostname you're testing")
@click.option('--follow-redirects', '-f', is_flag=True, default=False, help="Follow all redirects to get the real page")
@click.argument('domains', nargs=-1)
def test_response(ssl, www, user, password, domains, permutate, verbose, follow_redirects):
    "Test response from servers"
    protocols = ['http']
    if ssl:
        protocols += ['https']
    hostnames = get_hostnames(domains, permutate, www)
    for hostname in hostnames:
        for proto in protocols:
            url = u"{proto}://{hostname}".format(**locals())
            test_url(url, user, password, verbose, follow_redirects=follow_redirects)


def get_hostnames(domains, permutate='', www=True):
    if isinstance(domains, basestring):
        domains = [domains]
    if www:
        hostnames =  list(domains) + ['www.%s' % domain for domain in domains]
    else:
        hostnames = domains
    if permutate:
        hostnames = ['{}.{}'.format(host, ext) for host in hostnames for ext in permutate.split()]
    return hostnames


def test_url(url, user=None, password=None, verbose=False, follow_redirects=False):
    if verbose:
        click.secho("*** %s" % url, fg='magenta')
    if user:
        auth = HTTPBasicAuth(user, password)
    else:
        auth = None
    try:
        resp = requests.get(url, auth=auth, allow_redirects=follow_redirects)
    except requests.exceptions.ConnectionError:
        click.secho(u"{}: connection FAILED".format(url), fg='red')
        return

    if resp.status_code >= 300 and resp.status_code < 400:
        Location = resp.headers['Location']
        msg = u"{url} => {resp.status_code}  ({Location})".format(**locals())
        click.secho(msg)
    elif resp.status_code >= 200 and resp.status_code < 300:
        d = pyquery.PyQuery(resp.content)
        try:
            title = (d('title').contents()[0]).strip()
        except IndexError:
            title = "NO TITLE"

        msg = u"{url} => {resp.status_code} ({title})".format(**locals())
        click.echo(click.style(msg, fg='green'))
    else:
        msg = u"{url} ==> {resp.status_code} {resp.reason}".format(url=url, resp=resp)
        click.echo(click.style(msg, fg='red'))

# https://serverlesscode.com/post/ssl-expiration-alerts-with-lambda/

@click.command('expiry')
@click.option('--number', '-n', is_flag=True, help="Return just number", default=False)
@click.option('--www/--no-www', is_flag=True, help="Test/Don't prepend www", default=False)
@click.option('--permutate', '-x', help="Permutate on all domains in string eg: -x 'it com org'")
@click.option('--buffer-days', '-d', help="max days to issue a warning", default=BUFFER_DAYS)
#@click.option('--follow-redirects', '-f', is_flag=True, default=False, help="Follow all redirects to get the real page")
@click.argument('domains', nargs=-1)
def ssl_expires_in(domains, buffer_days, permutate, www, number):
    """Check hostname certificate expiry. """

    domains = get_hostnames(domains, permutate, www)
    for domain in domains:

        remaining = ssl_valid_time_remaining(domain)
        analyze_expiry(domain, remaining, number=number, buffer_days=buffer_days)


def analyze_expiry(domain, remaining, number=False, buffer_days=None):
    # if the cert expires in less than two weeks, we should reissue it
    if remaining < datetime.timedelta(days=0):
        # cert has already expired - uhoh!
        if number:
            click.echo('-1')
        else:
            click.secho(u"{}: Already expired".format(domain), fg='red')
        #raise AlreadyExpired("Cert expired %s days ago" % remaining.days)
    elif remaining < datetime.timedelta(days=buffer_days):
        # expires sooner than the buffer
        if number:
            click.echo(remaining.days)
        else:
            click.secho(u"{}: {} needs refresh".format(domain, remaining.days), fg='yellow')
    else:
        # everything is fine
        if number:
            click.echo(remaining.days)
        else:
            click.secho(u"{}: {} OK".format(domain, remaining.days), fg='green')
        #return False

def ssl_expiry_datetime(hostname):
    ssl_date_fmt = r'%b %d %H:%M:%S %Y %Z'

    context = ssl.create_default_context()
    conn = context.wrap_socket(
        socket.socket(socket.AF_INET),
        server_hostname=hostname,
    )
    # 3 second timeout because Lambda has runtime limitations
    conn.settimeout(3.0)

    conn.connect((hostname, 443))
    ssl_info = conn.getpeercert()
    # parse the string from the certificate into a Python datetime object
    return datetime.datetime.strptime(ssl_info['notAfter'], ssl_date_fmt)


def ssl_valid_time_remaining(hostname):
    """Get the number of days left in a cert's lifetime."""
    expires = ssl_expiry_datetime(hostname)
    return expires - datetime.datetime.utcnow()


@click.command('cert')
@click.option('--domain', '-d', is_flag=False, help="args are domain", default=False)
@click.option('--enabled', '-e', is_flag=True, help="look for certificates in nginx enabled ",
              default=False)
@click.option('--number', '-n', is_flag=True, help="Return just number", default=False)
@click.option('--buffer-days', '-d', help="max days to issue a warning", default=BUFFER_DAYS)
@click.option('--nginx-conf', '-N', help="nginx enabled base (def: {})".format(NGINX_ENABLED_DIR),
              default=NGINX_ENABLED_DIR)
@click.option('--letsencrypt-conf', '-l', default=LETSENCRYPT_CONF_DIR,
              help="nginx enabled base (def: {})".format(LETSENCRYPT_CONF_DIR),
              )
@click.argument('certificates', nargs=-1)
def get_expiry_from_cert(certificates, domain, number, buffer_days, enabled, letsencrypt_conf,
                         nginx_conf,):
    "Read expiry from file"

    if enabled:
        certificates = list(get_enabled_certificates(nginx_conf))

    if not domain:
        cert_files = certificates
    else:
        cert_files = []
        domains = certificates
        for domain in domains:
            cert_files += '/etc/letsencrypt/live/{}/cert.pem'.format(domain)
    for cert_file in cert_files:
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, open(cert_file).read())
        expiry = datetime.datetime.strptime(cert.get_notAfter(),"%Y%m%d%H%M%SZ")
        remaining = expiry - datetime.datetime.utcnow()
        domain = cert.get_subject().get_components()[0][1]
        analyze_expiry(domain, remaining, number, buffer_days)
        click.secho(u"    => {}".format(get_alternatives_names(cert_file)), fg='blue')

def get_alternatives_names(cert_file):
    cmd = 'openssl x509 -in {} -noout -text'.format(cert_file)
    p = sp.Popen(cmd.split(), stdin=sp.PIPE, stdout=sp.PIPE)
    txt = p.stdout.read()
    for line in txt.splitlines():
        if 'DNS' in line:
            return line.strip().replace('DNS:', '')


def get_enabled_certificates(nginx_conf):
    "get all certificates declared in nginx enabled configurations"

    for conf in glob.glob(os.path.join(nginx_conf, '*')):
        with open(conf) as f:
            for line in f:
                m = re.match('\s*ssl_certificate\s+(?P<cert>[\w\./]+)', line)
                if m:
                    yield m.group('cert')
